import shutil
import subprocess
from pathlib import Path
from typing import Mapping

import pytest

from bldr.bldr import BLDR


@pytest.fixture
def git_project_path(tmp_path: Path, asset_dir: Path, git_env: Mapping[str, str]) -> Path:
    git_project_dir = tmp_path.joinpath('git_project')
    shutil.copytree(
        asset_dir.joinpath('test-git-proj'),
        git_project_dir,
    )

    subprocess.check_call(['git', 'init'], cwd=git_project_dir)
    subprocess.check_call(['git', 'remote', 'add', 'origin', 'http://example.com/test-git-proj.git'], cwd=git_project_dir)
    subprocess.check_call(['git', 'add', '--all'], cwd=git_project_dir)
    subprocess.check_call(['git', 'commit', '--no-verify', '--message', 'Initial commit for production use'], cwd=git_project_dir, env=git_env)

    return git_project_dir


def test_git_project_debian_build(local_repo_dir: Path, git_project_path: Path, docker_from: str):
    bldr = BLDR(
        local_repo_dir=local_repo_dir,
        source_dir=git_project_path.parent,
        docker_from=docker_from,
    )
    bldr.build(git_project_path)

    dsc_file_path = local_repo_dir.joinpath('git_project', 'debs', 'git-proj_2.3-4.dsc')
    assert dsc_file_path, "Built dsc file should be in the local repo directory"

    dsc_file_content = dsc_file_path.read_text()
    assert "Vcs-Git: http://example.com/test-git-proj.git" in dsc_file_content

    commit_id = subprocess.check_output(['git', 'rev-parse', 'HEAD'], cwd=git_project_path).decode('utf-8')
    assert "Vcs-Git-Commit-Id: {}".format(commit_id) in dsc_file_content



import logging
import os
import pwd
import re
import shutil
import sys
from logging import Logger
from pathlib import Path
from typing import Dict, IO, List, Union, Optional
from tempfile import TemporaryDirectory

from .docker_utils import DockerImageBuilder, DockerImage, DockerContainer
from .utils import BLDRError, BLDRSetupFailed, escape_docker_image_tag, get_resource


PRE_BUILD_HOOK = "/hooks/pre-build"


class BLDR:

    def __init__(self,
        local_repo_dir: Path,
        source_dir: Path,
        docker_from: str,
        deb_build_options: Optional[str] = None,
        debug_shell: bool = False,
        snapshot: bool = False,
        nocache: bool = False,
        logger: Logger = logging.getLogger("bldr"),
        container_env: Optional[Dict] = None,
        hooks_dir: Optional[Path] = None,
        disable_tmpfs: bool = False,
    ) -> None:

        if ("\n" in docker_from or " " in docker_from):
            raise BLDRSetupFailed("Invalid docker_from parameter: {!r}".format(docker_from))
        self._docker_from = docker_from

        self._local_repo_dir = local_repo_dir.resolve()
        self._source_dir = source_dir.resolve()

        self._debug_shell = debug_shell
        self._deb_build_options = deb_build_options
        self._snapshot = snapshot
        self._nocache = nocache

        if container_env is None:
            self._container_env = {}
        else:
            self._container_env = container_env.copy()

        self._hooks_dir = hooks_dir
        self._logger = logger

        self._nonpriv_user_uid = self._get_nonpriv_user_uid()
        self._nonpriv_user_name = pwd.getpwuid(self._nonpriv_user_uid).pw_name

        self._tmp_on_tmpfs = not disable_tmpfs

    @property
    def local_repo_dir(self) -> Path:
        return self._local_repo_dir

    def _get_nonpriv_user_uid(self) -> int:
        local_uid = os.getuid()
        if local_uid != 0:
            return local_uid

        nonpriv_user_uid = int(os.environ.get('SUDO_UID', default='0'))
        if nonpriv_user_uid != 0:
            return nonpriv_user_uid

        raise BLDRSetupFailed("Refusing to run as root. Please run this script as a non-root user.")

    @classmethod
    def _get_source_package_name(cls, package_dir: Path) -> str:
        control_file = package_dir.joinpath('debian', 'control')
        with control_file.open('r') as fp:
            for line in fp:
                if line.startswith('Source: '):
                    return line.split('Source: ', 1)[1].strip()

        raise BLDRSetupFailed("Cannot parse {}".format(control_file))

    @classmethod
    def _get_clean_package_name(cls, package_name: str) -> str:
        return re.sub('[^A-Za-z0-9_.-]+', '', package_name)

    def generate_dockerfile(self, template_file: Path, output_file: Path) -> None:
        template = template_file.read_text()
        content = template.format(docker_from=self._docker_from)
        output_file.write_text(content)

    def _build_image(self, tag: str, control_file: Optional[Path] = None) -> DockerImage:
        self._logger.debug('Start building container image.')
        buildargs = {}

        if os.environ.get('http_proxy', None) is not None:
            buildargs['http_proxy'] = os.environ.get('http_proxy')
        if os.environ.get('no_proxy', None) is not None:
            buildargs['no_proxy'] = os.environ.get('no_proxy')
        buildargs['NONPRIV_USER_UID'] = str(self._nonpriv_user_uid)
        buildargs['NONPRIV_USER_NAME'] = self._nonpriv_user_name

        with TemporaryDirectory(prefix="bldr_docker_dir_") as tmp_dir:
            docker_files_dir = Path(tmp_dir).joinpath('docker_files')
            shutil.copytree(str(get_resource('.')), str(docker_files_dir))
            if control_file is None:
                docker_files_dir.joinpath('control').write_text('')
            else:
                shutil.copyfile(str(control_file), str(docker_files_dir.joinpath('control')))

            dockerfile = docker_files_dir.joinpath('Dockerfile')
            self.generate_dockerfile(
                template_file=docker_files_dir.joinpath('Dockerfile.tpl'),
                output_file=dockerfile,
            )

            if self._hooks_dir is not None:
                docker_hooks_dir = docker_files_dir.joinpath("hooks")
                shutil.rmtree(str(docker_hooks_dir))
                shutil.copytree(str(self._hooks_dir), str(docker_hooks_dir))

            image = DockerImageBuilder().build(
                path=docker_files_dir,
                dockerfile=dockerfile.name,
                tag=tag,
                nocache=self._nocache,
                buildargs=buildargs,
            )

        return image

    def _create_container(self, image: DockerImage, command: Union[str, list, None] = None) -> DockerContainer:
        if command is None:
            command = ['/bin/sleep', '24h']    # with this command the containers stops surely after a day

        env_vars = {}
        if os.environ.get('http_proxy', None) is not None:
            env_vars['http_proxy'] = os.environ.get('http_proxy')
        if os.environ.get('no_proxy', None) is not None:
            env_vars['no_proxy'] = os.environ.get('no_proxy')
        if self._deb_build_options is not None:
            env_vars['DEB_BUILD_OPTIONS'] = self._deb_build_options
        env_vars['BLDR_SNAPSHOT'] = '1' if self._snapshot else ''
        env_vars['TERM'] = os.environ.get('TERM', 'xterm-256color')
        env_vars['USER'] = self._nonpriv_user_name

        for env_key, env_value in self._container_env.items():
            env_vars[env_key] = env_value

        self._local_repo_dir.mkdir(parents=True, exist_ok=True)

        container = image.create_container(
            command=command,
            environment=env_vars,
            user=self._nonpriv_user_uid,
            volumes={
                self._source_dir: {'bind': '/source', 'mode': 'z'},
                self._local_repo_dir: {'bind': '/local-apt', 'mode': 'z'}
            },
            tmp_on_tmpfs=self._tmp_on_tmpfs,
        )

        return container

    def get_package_relative_dir(self, package_dir: Path) -> Path:
        return package_dir.relative_to(self._source_dir)

    def _container_exec(self, container: DockerContainer, command: Union[str, List]):
        exitcode = container.exec(command)
        if self._debug_shell and exitcode > 0:
            container.exec_with_pty(command='/bin/bash')
        return exitcode

    def build(self, package_dir: Path) -> DockerImage:
        package_dir = package_dir.resolve()

        source_package_name = self._get_source_package_name(package_dir)
        clean_package_name = self._get_clean_package_name(source_package_name)

        tag = 'bldr-deb-builder-{release}:{source_package_name}'.format(
            release=escape_docker_image_tag(self._docker_from),
            source_package_name=clean_package_name,
        )

        image = self._build_image(tag=tag, control_file=package_dir.joinpath('debian', 'control'))
        with self._create_container(image) as container:
            exitcode = self._container_exec(container, command=[PRE_BUILD_HOOK])
            if exitcode > 0:
                raise BLDRError('BLDR pre-build hook failed with exit code {exitcode}'.format(exitcode=exitcode), exitcode)

            exitcode = self._container_exec(container, command=['build-deb', str(self.get_package_relative_dir(package_dir))])
            if exitcode > 0:
                raise BLDRError('BLDR build failed with exit code {exitcode}'.format(exitcode=exitcode), exitcode)

        return image

    @classmethod
    def selftest(cls) -> None:
        for ubuntu_release in ['xenial', 'bionic', 'focal']:
            image = DockerImage(image='ubuntu:{}'.format(ubuntu_release))
            assert image is not None, "DockerImage should be initialized without Exception"

            with image.create_container(command=['sleep', '3600']) as container:
                assert container is not None, "docker should be able to start an ubuntu {} container".format(ubuntu_release)

                assert container.exec(command=['false']) == 1, "return code should be returned"

                cmd = 'useradd -u 1000 user && su user -c whoami'   # checking for possible SELinux issues
                assert container.exec_run(command=['bash', '-c', cmd]) == 'user\n', "'su user' should succeed"

                cmd = 'echo "deb http://archive.ubuntu.com/ubuntu {} main universe" > /etc/apt/sources.list'.format(ubuntu_release)
                container.exec_run(command=['bash', '-c', cmd])

                cmd = 'export DEBIAN_FRONTEND=noninteractive; apt-get -qq update && apt-get -qq install zip >/dev/null'
                container.exec_run(command=['bash', '-c', cmd])

    def reindex(self) -> None:
        tag = 'bldr-deb-builder-{release}:reindex'.format(
            release=escape_docker_image_tag(self._docker_from),
        )

        image = self._build_image(tag=tag)
        with self._create_container(image) as container:
            exitcode = self._container_exec(container, command=['reindex'])
            if exitcode > 0:
                raise BLDRError('BLDR reindex failed with exit code {exitcode}'.format(exitcode=exitcode), exitcode)

    def shell(self, package_dir: Path, stdin: IO[str] = None) -> None:
        package_dir = package_dir.resolve()

        control_file = package_dir.joinpath('debian', 'control')
        if control_file.is_file():
            source_package_name = self._get_source_package_name(package_dir)
            clean_package_name = self._get_clean_package_name(source_package_name)
            suffix = clean_package_name
        else:
            control_file = None     # type: ignore
            suffix = 'shell'

        tag = 'bldr-deb-builder-{release}:{suffix}'.format(
            release=escape_docker_image_tag(self._docker_from),
            suffix=suffix,
        )

        image = self._build_image(tag=tag, control_file=control_file)
        if os.isatty(0) and not stdin:
            container = self._create_container(image=image, command=['/bin/bash'])
            exitcode = container.run_with_pty(interactive=True)
        else:
            if stdin is None:
                stdin = sys.stdin
            command = stdin.read()
            container = self._create_container(image=image, command=['/bin/bash', '-c', command])
            exitcode = container.run_with_pty()

        if exitcode > 0:
            raise BLDRError('BLDR shell failed with exit code {exitcode}'.format(exitcode=exitcode), exitcode)



import argparse
import logging
import os
import sys
import json
from pathlib import Path
from textwrap import dedent
from typing import List, Tuple

from .bldr import BLDR
from .version import get_version
from .utils import BLDRError, escape_docker_image_tag, get_config_file_paths
from .config import ArgumentParser, JSONConfigLoader, ConfigLoader, SubParsers


EXTRA_HELP = """
There are 2 repositories used in building, in priority order:\n
    1. The local repository, to which the built packages go. This is here to enable building packages that depend on
        other locally built packages.\n
    2. The pre-configured repositories\n
\n
The packages will be installed from the highest priority repository where they are available, even if they are older
than the ones in the following repositories where it exists, with the version corresponding to that repository.\n
\n
Important info: -dev packages usually depend on the normal ones with a "= version" clause. If an old package is present
in a high priority repository (such as the filtering proxy), but the -dev package is not, the -dev one will be acquired
from a lower priority repo, possibly with a different version. This will cause apt to fail with a dependency error,
because it can't resolve such situations.
"""

logging.basicConfig(format="%(message)s", level=logging.DEBUG, stream=sys.stdout)
logging.getLogger("docker").setLevel(logging.WARNING)
logging.getLogger("urllib3").setLevel(logging.WARNING)


class CLI:
    description = ""

    def __init__(self, argv: List[str], logger: logging.Logger = logging.getLogger("cli")) -> None:
        self._logger = logger
        self.args = self._parse_arguments(argv)
        if hasattr(self.args, 'container_env'):
            self.args.container_env = self.format_container_env(
                self.args.container_env
            )

    def format_container_env(self, elements: List[str]) -> List[Tuple[str, str]]:
        retval = []
        for element in elements:
            if isinstance(element, str):
                key, value = element.split("=", 1)
                retval.append((key, value))
            if isinstance(element, (tuple, list)):
                retval.append(element)
        return retval

    @classmethod
    def _add_common_arguments(cls, parser: argparse.ArgumentParser) -> None:

        parser.add_argument(
            "docker_from",
            help="Specify the value which will be used for the FROM keyword when building the docker image. Example: 'ubuntu:bionic'",
        )
        parser.add_argument(
            '-s', '--snapshot',
            help="Build a snapshot version. "
                 "If this flag is set, a dummy entry is added to the debian changelog file, "
                 "and +xsnapshot+$date will be appended to the package version.",
            action='store_true',
        )
        parser.add_argument(
            '--shell',
            help="Launch a debug shell in the running docker container if build fails."
                 "(default: %(default)s)",
            action='store_true',
        )
        parser.add_argument(
            '--deb-build-options',
            help="This variable can contain several flags to change how a package is compiled and built. "
                 "Each flag must be in the form flag or flag=options. "
                 "If multiple flags are given, they must be separated by whitespace. "
                 "(default: %(default)s)",
            action='store', default=os.environ.get('DEB_BUILD_OPTIONS', None)
        )
        parser.add_argument(
            '--local-repo-dir',
            help="The directory for the local apt repository. The value will be: "
                 "1. This argument, if given. "
                 "2. $BLDR_LOCAL_REPO_DIR environment variable, if set. "
                 "3. 'local-apt-<image_tag>, in the cwd. Image tag will be normalized, all ':' and '/' will be replaced to '-'",
            action='store',
            type=Path,
        )
        parser.add_argument(
            '--nocache',
            help="Disable caching of docker image layers. "
                 "Useful to sanity-check that an error is not caused by cache layers. "
                 "A dedicated CI machine should use this option. "
                 "(default: %(default)s)",
            action='store_true'
        )
        parser.add_argument(
            "--container-env",
            help="Specify environment variables for the container in the form of key=value. Can be specified multiple times.",
            action="append",
            default=[],
        )
        parser.add_argument(
            "--hooks-dir",
            help="A path pointing to a directory containing the hooks",
            type=Path,
        )
        parser.add_argument(
            "--disable-tmpfs",
            help="Disable using tmpfs.",
            action="store_true",
        )

    def _create_config_loader(self) -> ConfigLoader:
        return JSONConfigLoader(get_config_file_paths())

    def _parse_arguments(self, argv: List) -> argparse.Namespace:
        config_loader = self._create_config_loader()
        parser = ArgumentParser(
            description=self.description,
            epilog=self._extra_help,
            formatter_class=argparse.RawTextHelpFormatter,
            config_loader=config_loader,
        )
        parser.add_argument(
            '-v', '--version',
            action='version', version='%(prog)s {version}'.format(version=get_version())
        )
        parser.add_argument(
            '--config',
            help='JSON File used to set command line parameter default values, in addition to the defaults written below.',
            metavar='PATH',
            type=Path,
        )

        subparsers = SubParsers(config_loader, parser.add_subparsers(metavar='command', help='description'))

        build_option_parser = subparsers.add_parser(
            'build',
            help="This command builds snapshot .deb files in a docker container."
                 "Call from any source dir that has a debian directory, from just outside the debian directory."
                 "Places compiled .deb files into the specified repository.",
        )
        build_option_parser.add_argument(
            'package',
            nargs='?', default=Path.cwd(),
            help="Source directory, which will be built. (default: %(default)s)",
            type=Path,
        )
        build_option_parser.set_defaults(action=lambda self, args: self.build(args.package))
        self._add_common_arguments(build_option_parser)

        reindex_option_parser = subparsers.add_parser(
            'reindex',
            help="This command reindexes a simple apt-repository (i.e. creates Packages and Sources files)."
                 "The working directory is arbitrary, if BLDR_LOCAL_REPO_DIR is set "
                 "or --local-repo-dir is specified via command line.",
        )
        reindex_option_parser.set_defaults(action=lambda self, args: self.reindex())
        self._add_common_arguments(reindex_option_parser)

        selftest_option_parser = subparsers.add_parser(
            'selftest',
            help="This command runs system configuration tests (e.g. is docker properly configured "
                 "and works for all supported Ubuntu releases).",
        )
        selftest_option_parser.set_defaults(action=lambda self, args: self.selftest())

        shell_option_parser = subparsers.add_parser(
            'shell',
            help="This command puts you in a docker shell, which already contains the build dependencies of your project"
                 "Call from any source dir that has a debian directory, from just outside the debian directory."
                 "It mounts your home directory into the docker container.",
        )
        shell_option_parser.add_argument(
            'package',
            nargs='?', default=Path.cwd(),
            help="Source directory, where the dependencies come from. (default: %(default)s)",
            type=Path,
        )
        shell_option_parser.set_defaults(action=lambda self, args: self.shell(args.package))
        self._add_common_arguments(shell_option_parser)

        ns, _ = parser.parse_known_args(argv[1:])
        if ns.config:
            try:
                config_loader.load_config_file(ns.config)
            except OSError:
                parser.error("Unable to open configuration file: '{}'".format(ns.config))
            except json.JSONDecodeError as err:
                parser.error("Unable to parse configuration file: '{}'".format(err))
            parser.config_loader.paths.append(ns.config)

        return parser.parse_args(argv[1:])

    @property
    def _extra_help(self) -> str:
        config_loader = self._create_config_loader()
        config_extra_help = dedent("""

            Configuration files:

              You can use configuration files to change the default settings.
              The format of the config file is JSON-formatted dictionary. The keys are the parameter names, but '-' characters are replaced with '_'.
              The order of looking up a parameter value is:
                - Command line argument
                - If --config is set, read from that file
                - Search in default config files (in order):
        """)
        for config_file_path in reversed(config_loader.paths):
            config_extra_help += "      {}\n".format(config_file_path)
        config_extra_help += "    - Default value"

        return EXTRA_HELP + config_extra_help

    def _find_local_repo_dir(self) -> Path:
        local_repo_dir = self.args.local_repo_dir

        if not local_repo_dir:
            local_repo_dir = os.environ.get("BLDR_LOCAL_REPO_DIR")
        if not local_repo_dir:
            local_repo_dir = Path().cwd().joinpath("local-apt-{}".format(escape_docker_image_tag(self.args.docker_from)))

        return local_repo_dir

    def _init_bldr(self) -> BLDR:
        bldr = BLDR(
            local_repo_dir=self._find_local_repo_dir(),
            source_dir=Path().cwd().parent,
            docker_from=self.args.docker_from,
            deb_build_options=self.args.deb_build_options,
            debug_shell=self.args.shell,
            snapshot=self.args.snapshot,
            nocache=self.args.nocache,
            container_env=dict(self.args.container_env),
            hooks_dir=self.args.hooks_dir,
            disable_tmpfs=self.args.disable_tmpfs,
        )
        return bldr

    def main(self) -> int:
        try:
            self.args.action(self, self.args)
        except BLDRError as bldr_err:
            self._logger.error(bldr_err.msg)
            return bldr_err.exitcode

        return 0

    def build(self, package_dir: Path) -> None:
        self._logger.debug("BLDR build with {}".format(package_dir))

        bldr = self._init_bldr()
        bldr.build(package_dir=package_dir)
        deb_path = bldr.local_repo_dir.joinpath(bldr.get_package_relative_dir(package_dir), 'debs')
        self._logger.debug("Your deb files are at {}".format(deb_path))

    def reindex(self) -> None:
        self._logger.debug("BLDR reindex")

        bldr = self._init_bldr()
        bldr.reindex()

    def selftest(self) -> None:
        self._logger.debug("BLDR selftest")

        BLDR.selftest()

        self._logger.debug("Selftest: PASSED")

    def shell(self, package_dir: Path) -> None:
        self._logger.debug("BLDR shell")

        bldr = self._init_bldr()
        bldr.shell(package_dir=package_dir)


def main(argv: List[str] = sys.argv) -> None:
    cli = CLI(argv=argv)
    sys.exit(cli.main())


if __name__ == '__main__':
    main()




import os
import pytest


@pytest.fixture()
def git_env():
    env = os.environ.copy()
    env["GIT_COMMITTER_NAME"] = "Test Developer"
    env["GIT_COMMITTER_EMAIL"] = "test.developer@example.com"
    env["GIT_AUTHOR_NAME"] = env["GIT_COMMITTER_NAME"]
    env["GIT_AUTHOR_EMAIL"] = env["GIT_COMMITTER_NAME"]

    yield env



import os
import pwd
from pathlib import Path
from pkg_resources import resource_filename


class BLDRError(Exception):
    def __init__(self, msg: str, exitcode: int = 1) -> None:
        self.msg = msg
        self.exitcode = exitcode

    def __str__(self) -> str:
        return self.msg


class BLDRSetupFailed(BLDRError):
    def __init__(self, msg: str, exitcode: int = 1) -> None:
        msg = 'Setting up BLDR failed: {}'.format(msg)
        super().__init__(msg, exitcode)


def get_resource(path: str) -> Path:
    return Path(resource_filename('bldr', str(Path('data', path))))


def escape_docker_image_tag(tag: str) -> str:
    return tag.replace(":", "-").replace("/", "-")


def get_home_dir() -> Path:
    if 'SUDO_UID' in os.environ and os.geteuid() == 0:
        uid = int(os.environ['SUDO_UID'])
    else:
        uid = os.getuid()

    pw_entry = pwd.getpwuid(uid)
    home = Path(pw_entry.pw_dir)

    return home


def get_config_file_paths():
    home_dir = get_home_dir()
    config_file_paths = [
        Path('/etc/bldr.conf'),
        home_dir.joinpath('.config/bldr.conf'),
        home_dir.joinpath('.bldr.conf'),
        Path('bldr.conf'),
    ]

    return config_file_paths



import logging
import os
import socket
from logging import Logger
from typing import Any, Dict, List, Optional, Union
from pathlib import Path

import docker
import dockerpty
from docker import DockerClient
from docker.models.images import Image
from docker.errors import APIError, DockerException
from requests import RequestException

from .utils import BLDRSetupFailed


def _create_docker_client() -> DockerClient:
    try:
        return docker.from_env(version='auto')
    except DockerException as e:
        raise BLDRSetupFailed(
            'Cannot create Docker client. Is Docker daemon running?\nAdditional info: {}'.format(e)
        )


def _check_docker_client(client: DockerClient) -> None:
    try:
        client.ping()
    except (DockerException, RequestException) as e:
        raise BLDRSetupFailed(
            'Cannot connect to Docker daemon. Is Docker daemon running?\nAdditional info: {}'.format(e)
        )


class DockerImageBuilder:
    def __init__(self, client: Optional[DockerClient] = None, logger: Logger = logging.getLogger('DockerImageBuilder')) -> None:
        self._logger: logging.Logger = logger

        if client is None:
            client = _create_docker_client()
        self._client: DockerClient = client
        _check_docker_client(self._client)

    def build(self, path: Path, dockerfile: str, tag: str, buildargs: Dict, nocache: bool = False) -> 'DockerImage':
        stream = self._client.api.build(
            path=str(path),
            dockerfile=dockerfile,
            tag=tag,
            forcerm=True,
            nocache=nocache,
            buildargs=buildargs,
            decode=True,
        )
        for chunk in stream:
            if chunk.get('stream', None) is not None:
                self._logger.debug(chunk.get('stream').strip())
            elif chunk.get('errorDetail', None) is not None:
                raise DockerException(chunk.get('error'))

        return DockerImage(client=self._client, image=tag)


class DockerImage:
    def __init__(self, image: Union[str, Image], client: Optional[DockerClient] = None, logger: Optional[Logger] = None) -> None:
        if client is None:
            client = _create_docker_client()
        self._client = client
        _check_docker_client(self._client)

        self._logger = logger
        if self._logger is None:
            self._logger = logging.getLogger('DockerImage')

        self._tag = image

    def create_container(self, **kwargs: Any) -> 'DockerContainer':
        return DockerContainer(client=self._client, image=self._tag, **kwargs)


class DockerContainer:
    def __init__(
        self,
        image: Union[str, Image],
        command: Union[str, List],
        environment: Optional[Dict] = None,
        user: Optional[str] = None,
        volumes: Optional[Dict] = None,
        client: Optional[DockerClient] = None,
        logger: Logger = logging.getLogger('DockerContainer'),
        tmp_on_tmpfs: bool = True,
    ) -> None:

        if client is None:
            client = _create_docker_client()
        self._client = client
        _check_docker_client(self._client)

        self._logger = logger

        try:
            self._client.images.get(image)
        except docker.errors.ImageNotFound:
            self._client.images.pull(image)

        tmpfs = {'/tmp': 'rw,exec'} if tmp_on_tmpfs else {}

        self._container = self._client.containers.create(
            init=True,
            image=image,
            command=command,
            stdin_open=True,
            tty=os.isatty(0),
            environment=environment,
            network='host',
            security_opt=['seccomp=unconfined'],
            tmpfs=tmpfs,
            user=user,
            volumes=volumes,
            extra_hosts={socket.gethostname(): "127.0.0.1"},
        )

    def __enter__(self) -> 'DockerContainer':
        self._container.start()
        return self

    def run_with_pty(self, interactive: bool = False) -> int:
        dockerpty.start(self._client.api, self._container.id, interactive=interactive, logs=True)
        exit_code = self.get_exit_code()
        self._container.remove()
        return exit_code

    def exec(self, command: Union[str, List]) -> int:
        exec_id = self._client.api.exec_create(container=self._container.id, cmd=command)
        stream = self._client.api.exec_start(exec_id=exec_id, stream=True)

        for chunk in stream:
            self._logger.debug(chunk.decode('utf-8', errors='ignore').strip())

        return self._client.api.exec_inspect(exec_id=exec_id).get('ExitCode', 0)

    def exec_run(self, command: Union[str, List]) -> str:
        exitcode, output = self._container.exec_run(command)
        if exitcode != 0:
            raise ValueError('The following command "{}" exited with code: {}'.format(command, exitcode))
        output = output.decode('utf-8', errors='ignore')
        return output

    def exec_with_pty(self, command: Union[str, List]) -> None:
        dockerpty.exec_command(self._client.api, self._container.id, command=command)

    def get_exit_code(self) -> int:
        return self._client.api.inspect_container(self._container.id)['State'].get('ExitCode', 0)

    def __exit__(self, exc_type, exc_val, exc_tb) -> None:
        try:
            self._container.kill()
        except APIError:
            pass
        finally:
            self._container.remove()



from pathlib import Path

from bldr.bldr import BLDR


def test_reindex(asset_dir: Path, reindex_data: Path, tmp_path: Path, docker_from: str):
    bldr = BLDR(
        local_repo_dir=reindex_data,
        source_dir=tmp_path,
        docker_from=docker_from,
    )
    bldr.reindex()

    expected_packages_content = asset_dir.joinpath('test-reindex-data', 'Packages.correct').read_text()
    assert reindex_data.joinpath('Packages').read_text() == expected_packages_content, 'Generated Packages file should be identical to Packages.correct'

    expected_sources_content = asset_dir.joinpath('test-reindex-data', 'Sources.correct').read_text()
    assert reindex_data.joinpath('Sources').read_text() == expected_sources_content, 'Generated Sources file should be identical to Sources.correct'



from pathlib import Path

version_file_path = Path(__file__).absolute().parent.joinpath('VERSION')

__version = None


def get_version() -> str:
    global __version

    if __version is None:
        with version_file_path.open('r') as version_file:
            __version = version_file.read().strip()

    return __version



import argparse
import json

from pathlib import Path

from bldr.config import ArgumentParser, JSONConfigLoader


EXAMPLE_CONFIG = """
{
    "foo": "bar",
    "some_param": 1234
}
    """

NO_SUCH_FILE_PATH = Path("/tmp/____no_such_file_____")


def test_argparse_full_config(tmp_path: Path):
    config_path = tmp_path.joinpath("config.json")
    config_path.open('w').write(EXAMPLE_CONFIG)

    parser = ArgumentParser(config_loader=JSONConfigLoader([config_path]))

    parser.add_argument("--foo", help="Foo", default="default")
    parser.add_argument("--some-param", help="Some parameter", type=int, default=12)

    ns = parser.parse_args([])
    assert ns.foo == "bar"
    assert ns.some_param == 1234


def test_argparse_full_config_override(tmp_path: Path):
    config_path = tmp_path.joinpath("config.json")
    config_path.open('w').write(EXAMPLE_CONFIG)

    parser = ArgumentParser(config_loader=JSONConfigLoader([config_path]))
    parser.add_argument("--foo", help="Foo", default="default")
    parser.add_argument("--some-param", help="Some parameter", type=int, default=12)

    ns = parser.parse_args(["--foo", "cmdline", "--some-param", "7777"])
    assert ns.foo == "cmdline"
    assert ns.some_param == 7777


def test_argparse_no_config():

    assert not NO_SUCH_FILE_PATH.is_file(), "{} should not exist".format(NO_SUCH_FILE_PATH)

    parser = ArgumentParser(config_loader=JSONConfigLoader([NO_SUCH_FILE_PATH]))
    parser.add_argument("--foo", help="Foo", default="default")
    parser.add_argument("--some-param", help="Some parameter", type=int, default=12)

    ns = parser.parse_args([])
    assert ns.foo == "default"
    assert ns.some_param == 12


def test_argparse_empty_config(tmp_path: Path):
    config_path = tmp_path.joinpath("config.json")
    config_path.open('w').write("{}")

    parser = ArgumentParser(config_loader=JSONConfigLoader([config_path]))
    parser.add_argument("--foo", help="Foo", default="default")
    parser.add_argument("--some-param", help="Some parameter", type=int, default=12)

    ns = parser.parse_args([])
    assert ns.foo == "default"
    assert ns.some_param == 12


def test_argparse_inheritance(tmp_path: Path):
    example_config = json.loads(EXAMPLE_CONFIG)

    config1_path = tmp_path.joinpath("config1.json")
    config1_path.open('w').write(EXAMPLE_CONFIG)

    example_config["foo"] = "config2"
    del example_config["some_param"]

    config2_path = tmp_path.joinpath("config2.json")
    config2_path.open('w').write(json.dumps(example_config))

    example_config["foo"] = "config3"
    example_config["some_param"] = None
    config3_path = tmp_path.joinpath("config3.json")
    config3_path.open('w').write(json.dumps(example_config))

    parser = ArgumentParser(config_loader=JSONConfigLoader([config1_path, config2_path, config3_path]))
    parser.add_argument("--foo", help="Foo", default="default")
    parser.add_argument("--some-param", help="Some parameter", type=int, default=12)

    ns = parser.parse_args([])

    assert ns.foo == "config3", "foo should come from config3.json"
    assert ns.some_param == 12, "some_param should come from argparse default as config3.json clears it"


def test_argparse_api_definitions():
    parser = argparse.ArgumentParser()
    assert hasattr(parser, "_actions"), "parser object should have '_actions' attribute"
    assert type(parser._actions) == list, "parser's '_actions' attribute must be a list"
    assert hasattr(argparse, "_StoreAction"), "argparse must have a '_StoreAction' class defined"



from pathlib import Path

from pytest_console_scripts import ScriptRunner

from bldr.version import get_version
from ..testutil import skip_if_global_config


def test_version(script_runner: ScriptRunner):
    ret = script_runner.run('bldr', '--version')
    assert ret.success, "version print should be succeded"
    assert ret.stdout == 'bldr {}\n'.format(get_version())
    assert ret.stderr == '', "nothing should be written to stderr"


@skip_if_global_config
def test_selftest(script_runner: ScriptRunner):
    ret = script_runner.run('bldr', 'selftest')
    assert ret.success, "selftest command should be succeded"
    assert ret.stderr == '', "nothing should be written to stderr"


@skip_if_global_config
def test_build(script_runner: ScriptRunner, asset_dir: Path, docker_from: str, local_repo_dir: Path):
    ret = script_runner.run('bldr', 'build', docker_from, "--local-repo-dir", str(local_repo_dir), cwd=asset_dir.joinpath('example-lang-C'))
    assert ret.success, "build command should be succeded"
    assert 'Your deb files are at' in ret.stdout
    assert ret.stderr == '', "nothing should be written to stderr"

    assert local_repo_dir.joinpath("Packages").is_file(), "Packages file should exist"
    assert local_repo_dir.joinpath("Sources").is_file(), "Packages file should exist"
    assert local_repo_dir.joinpath("Packages").stat().st_size > 0, "Packages file should be non-empty"
    assert local_repo_dir.joinpath("Sources").stat().st_size > 0, "Sources file should be non-empty"


@skip_if_global_config
def test_failed_build(script_runner: ScriptRunner, asset_dir: Path, docker_from: str, local_repo_dir: Path):
    ret = script_runner.run('bldr', 'build', docker_from, "--local-repo-dir", str(local_repo_dir), cwd=asset_dir.joinpath('never-builds'))
    assert not ret.success, "build command should be failed"
    assert ret.stderr == '', "nothing should be written to stderr"

    assert local_repo_dir.joinpath("Packages").is_file(), "Packages file should exist"
    assert local_repo_dir.joinpath("Sources").is_file(), "Sources file should exist"
    assert local_repo_dir.joinpath("Packages").stat().st_size == 0, "Packages file should be empty"
    assert local_repo_dir.joinpath("Sources").stat().st_size == 0, "Sources file should be empty"


@skip_if_global_config
def test_reindex(script_runner: ScriptRunner, asset_dir: Path, reindex_data: Path, tmp_path: Path, docker_from: str):
    ret = script_runner.run('bldr', 'reindex', docker_from, '--local-repo-dir', reindex_data)
    assert ret.success, "reindex command should be succeded"
    assert ret.stderr == '', "nothing should be written to stderr"

    expected_packages_content = asset_dir.joinpath('test-reindex-data', 'Packages.correct').read_text()
    assert reindex_data.joinpath('Packages').read_text() == expected_packages_content, 'Generated Packages file should be identical to Packages.correct'

    expected_sources_content = asset_dir.joinpath('test-reindex-data', 'Sources.correct').read_text()
    assert reindex_data.joinpath('Sources').read_text() == expected_sources_content, 'Generated Sources file should be identical to Sources.correct'


def test_command_names_in_help(script_runner: ScriptRunner):
    ret = script_runner.run('bldr', '--help')
    assert ret.success
    for command_name in ['build', 'shell', 'reindex', 'selftest']:
        assert command_name in ret.stdout


def test_command_help(script_runner: ScriptRunner):
    ret = script_runner.run('bldr', 'build', '--help')
    assert ret.success

    argument_list = [
        'docker_from',
        'package',
        '--snapshot',
        '--shell',
        '--deb-build-options',
        '--local-repo-dir',
        '--nocache',
        '--container-env',
        '--hooks-dir',
    ]
    for argument in argument_list:
        assert argument in ret.stdout


def test_config_with_a_non_existent_file(script_runner: ScriptRunner, tmp_path: Path):
    config = tmp_path.joinpath('non-existent-config.json')
    ret = script_runner.run(
        'bldr', '--config', str(config), 'build', 'ubuntu:bionic'
    )
    assert not ret.success
    assert "Unable to open configuration file: '{}'".format(config) in ret.stderr


def test_config_with_a_unreadable_file(script_runner: ScriptRunner, tmp_path: Path):
    config = tmp_path.joinpath('non-existent-config.json')
    config.touch()
    config.chmod(0)
    ret = script_runner.run(
        'bldr', '--config', str(config), 'build', 'ubuntu:bionic'
    )
    assert not ret.success
    assert "Unable to open configuration file: '{}'".format(config) in ret.stderr


def test_config_with_a_non_json_file(script_runner: ScriptRunner, tmp_path: Path):
    config = tmp_path.joinpath('non-existent-config.json')
    config.write_text('{"foo": ')

    ret = script_runner.run(
        'bldr', '--config', str(config), 'build', 'ubuntu:bionic'
    )
    assert not ret.success
    assert "Unable to parse configuration file: 'Expecting value: line 1 column 9 (char 8)'" in ret.stderr



import subprocess
import shutil
from pathlib import Path
from typing import Mapping

import pytest

from bldr.bldr import BLDR
from ..testutil import copytree, extract_deb


@pytest.fixture
def quilt_project_path(tmp_path: Path, asset_dir: Path, git_env: Mapping[str, str]) -> Path:
    quilt_project_dir = tmp_path.joinpath('quilt_project')
    quilt_project_dir.mkdir()
    subprocess.check_call(['git', 'init'], cwd=quilt_project_dir)
    subprocess.check_call(['git', 'checkout', '-b', 'upstream'], cwd=quilt_project_dir)

    copytree(asset_dir.joinpath('test-quilt-proj', 'upstream'), quilt_project_dir)
    subprocess.check_call(['git', 'add', '--all'], cwd=quilt_project_dir)
    subprocess.check_call(['git', 'commit', '--no-verify', '--message', 'Imported older upstream'], cwd=quilt_project_dir, env=git_env)
    subprocess.check_call(['git', 'tag', 'upstream-older'], cwd=quilt_project_dir)

    copytree(asset_dir.joinpath('test-quilt-proj', 'upstream-newer'), quilt_project_dir, exist_ok=True)
    subprocess.check_call(['git', 'add', '--all'], cwd=quilt_project_dir)
    subprocess.check_call(['git', 'commit', '--no-verify', '--message', 'Imported newer upstream'], cwd=quilt_project_dir, env=git_env)
    subprocess.check_call(['git', 'tag', 'upstream-newer'], cwd=quilt_project_dir)

    subprocess.check_call(['git', 'checkout', 'upstream-older'], cwd=quilt_project_dir)
    subprocess.check_call(['git', 'checkout', '-b', 'ubuntu'], cwd=quilt_project_dir)

    copytree(asset_dir.joinpath('test-quilt-proj', 'debian'), quilt_project_dir, exist_ok=True)
    subprocess.check_call(['git', 'add', '--all'], cwd=quilt_project_dir)
    subprocess.check_call(['git', 'commit', '--no-verify', '--message', 'Imported debian older version'], cwd=quilt_project_dir, env=git_env)
    subprocess.check_call(['git', 'tag', 'debian-older'], cwd=quilt_project_dir)

    subprocess.check_call(['git', 'merge', 'upstream-newer', '--no-commit'], cwd=quilt_project_dir, env=git_env)
    copytree(asset_dir.joinpath('test-quilt-proj', 'debian-newer'), quilt_project_dir, exist_ok=True)
    subprocess.check_call(['git', 'add', '--all'], cwd=quilt_project_dir)
    subprocess.check_call(['git', 'commit', '--no-verify', '--message', 'Imported debian newer version'], cwd=quilt_project_dir, env=git_env)
    subprocess.check_call(['git', 'tag', 'debian-newer'], cwd=quilt_project_dir)

    subprocess.check_call(['git', 'checkout', '-b', 'master'], cwd=quilt_project_dir)
    copytree(asset_dir.joinpath('test-quilt-proj', 'master'), quilt_project_dir, exist_ok=True)
    subprocess.check_call(['git', 'add', '--all'], cwd=quilt_project_dir)
    subprocess.check_call(['git', 'commit', '--no-verify', '--message', 'Our patches'], cwd=quilt_project_dir, env=git_env)

    return quilt_project_dir


def do_build(local_repo_dir: Path, quilt_project_path: Path, docker_from: str, tmp_path: Path, expected_output: str) -> None:
    bldr = BLDR(
        local_repo_dir=local_repo_dir,
        source_dir=quilt_project_path.parent,
        docker_from=docker_from,
    )
    bldr.build(quilt_project_path)

    quilt_proj_deb_file = list(local_repo_dir.glob('**/quilt-proj*.deb'))[0]

    extract_dir = tmp_path.joinpath('extracted')
    extract_dir.mkdir()
    extract_deb(quilt_proj_deb_file, extract_dir)

    output = subprocess.check_output(['usr/bin/am-i-quilted'], cwd=extract_dir).decode('utf-8')
    assert output == expected_output, "The script output should be what expected"

    shutil.rmtree(str(extract_dir), ignore_errors=True)


def test_quilt_project_build(local_repo_dir: Path, quilt_project_path: Path, docker_from: str, tmp_path: Path):
    subprocess.check_call(['git', 'checkout', 'debian-older'], cwd=quilt_project_path)
    expected_output = (
        "So I am a little script from upstream\n"
        "A little bit buggy though.\n"
        "But do you still love me?\n"
    )
    debian_older_local_repo_dir = local_repo_dir.joinpath('debian-older')
    do_build(debian_older_local_repo_dir, quilt_project_path, docker_from, tmp_path, expected_output)

    subprocess.check_call(['git', 'checkout', 'debian-newer'], cwd=quilt_project_path)
    expected_output = (
        "So I am a newer little script from upstream\n"
        "A little bit buggy though.\n"
        "But do you still love me?\n"
    )
    debian_newer_local_repo_dir = local_repo_dir.joinpath('debian-newer')
    do_build(debian_newer_local_repo_dir, quilt_project_path, docker_from, tmp_path, expected_output)

    subprocess.check_call(['git', 'checkout', 'master'], cwd=quilt_project_path)
    expected_output = (
        "So I am a newer little script from upstream\n"
        "A little bit buggy though.\n"
        "But do you still really love me?\n"
    )
    master_local_repo_dir = local_repo_dir.joinpath('master')
    do_build(master_local_repo_dir, quilt_project_path, docker_from, tmp_path, expected_output)

    debian_newer_dsc = debian_newer_local_repo_dir.joinpath('quilt_project', 'debs', 'quilt-proj_1.1-1ubuntu1.dsc')
    master_dsc = master_local_repo_dir.joinpath('quilt_project', 'debs', 'quilt-proj_1.1-1ubuntu1bb50.1.dsc')

    assert 'quilt-proj_1.1.orig.tar.gz' in debian_newer_dsc.read_text(), "The script should generate the same orig.tar.gz every time."
    assert 'quilt-proj_1.1.orig.tar.gz' in master_dsc.read_text(), "The script should generate the same orig.tar.gz every time."



import pytest
import re
from pathlib import Path
from typing import List

from bldr.bldr import BLDR, BLDRError
from ..testutil import extract_deb


def get_local_deb_files(local_repo_dir: Path, package_prefix: str = '') -> List[str]:
    deb_file_paths = list(local_repo_dir.glob('**/{}*.deb'.format(package_prefix)))
    assert deb_file_paths, "Built debian package should be in the local repo directory"
    return [str(Path('/local-apt').joinpath(deb_file_path.relative_to(local_repo_dir))) for deb_file_path in deb_file_paths]


def test_debian_build_lang_c(local_repo_dir: Path, asset_dir: Path, docker_from: str):
    bldr = BLDR(
        local_repo_dir=local_repo_dir,
        source_dir=asset_dir,
        docker_from=docker_from,
    )
    image = bldr.build(asset_dir.joinpath('example-lang-C'))

    local_deb_files = get_local_deb_files(local_repo_dir, package_prefix='dummy_1.0-1_amd64')
    assert len(local_deb_files) == 1, "one debian package should match to the glob pattern"

    with bldr._create_container(image) as container:
        output = container.exec_run(
            command=['dpkg', '--contents', local_deb_files[0]]
        )
        deb_content = '\n'.join(sorted([line.split()[5] for line in output.strip().split('\n')]))
        expected_content = asset_dir.joinpath('expected.example-lang-c.contents').read_text()
        assert deb_content == expected_content, "Deb file should contain the expected files"

        output = container.exec_run(
            command=['dpkg', '--field', local_deb_files[0]]
        )
        assert 'Package: dummy' in output, "package name should be in the debian fields"
        assert "Version: 1.0-1" in output, "version field should be in the debian fields in the given format"
        assert 'Maintainer: Test Maintainer <test.maintainer@example.com>' in output, "maintaner should be in the debian fields"


def test_snapshot_version(local_repo_dir: Path, asset_dir: Path, docker_from: str):
    bldr = BLDR(
        local_repo_dir=local_repo_dir,
        source_dir=asset_dir,
        docker_from=docker_from,
        snapshot=True,
    )
    image = bldr.build(asset_dir.joinpath('example-lang-C'))

    with bldr._create_container(image) as container:
        local_deb_files = get_local_deb_files(local_repo_dir, package_prefix='dummy_1.0-1+xsnapshot')
        assert len(local_deb_files) == 1, "one debian package should match to the glob pattern"

        output = container.exec_run(
            command=['dpkg', '--contents', local_deb_files[0]]
        )
        deb_content = '\n'.join(sorted([line.split()[5] for line in output.strip().split('\n')]))
        expected_content = asset_dir.joinpath('expected.example-lang-c.contents').read_text()
        assert deb_content == expected_content, "Deb file should contain the expected files"

        output = container.exec_run(
            command=['dpkg', '--field', local_deb_files[0]]
        )
        assert 'Package: dummy' in output, "package name should be in the debian fields"
        assert re.search(r"^Version: 1\.0-1\+xsnapshot\+\d{4}(?:\.\d{2}){5}\+1$", output, flags=re.MULTILINE), "version field should be in the debian fields in the given format"
        assert 'Maintainer: Test Maintainer <test.maintainer@example.com>' in output, "maintaner should be in the debian fields"


def test_depencies_are_satisfied_using_local_apt(local_repo_dir: Path, asset_dir: Path, docker_from: str):
    bldr = BLDR(
        local_repo_dir=local_repo_dir,
        source_dir=asset_dir,
        docker_from=docker_from,
    )
    bldr.build(asset_dir.joinpath('libdependable'))
    image = bldr.build(asset_dir.joinpath('depend-test'))

    with bldr._create_container(image) as container:
        local_deb_files = get_local_deb_files(local_repo_dir)
        container.exec_run(['sudo', 'dpkg', '-i'] + local_deb_files)
        output = container.exec_run(['run-depend-test'])
        assert re.search(r"^I am", output, flags=re.MULTILINE), "run-depend-test output should contain the given string"


def test_symlinks(local_repo_dir: Path, asset_dir: Path, docker_from: str):
    bldr = BLDR(
        local_repo_dir=local_repo_dir,
        source_dir=asset_dir,
        docker_from=docker_from,
    )
    bldr.build(asset_dir.joinpath('symlinks'))


def test_build_fails(local_repo_dir: Path, asset_dir: Path, docker_from: str):
    bldr = BLDR(
        local_repo_dir=local_repo_dir,
        source_dir=asset_dir,
        docker_from=docker_from,
    )
    with pytest.raises(BLDRError):
        bldr.build(asset_dir.joinpath('never-builds'))
        pytest.fail("SBS should raise exception, when an error occurs during build.")


def test_debian_build_with_hooks(local_repo_dir: Path, asset_dir: Path, docker_from: str):
    bldr = BLDR(
        local_repo_dir=local_repo_dir,
        source_dir=asset_dir,
        docker_from=docker_from,
        hooks_dir=asset_dir.joinpath('test-hooks')
    )
    bldr.build(asset_dir.joinpath('example-lang-C'))

    assert local_repo_dir.joinpath('pre-build.called').exists(), "pre-build hook should be called"
    assert local_repo_dir.joinpath('pre-init.called').exists(), "pre-init hook should be called"
    assert local_repo_dir.joinpath('pre-install-deps.called').exists(), "pre-install-deps hook should be called"
    assert local_repo_dir.joinpath('post-install-deps.called').exists(), "post-install-deps hook should be called"


def test_deb_build_options(local_repo_dir: Path, asset_dir: Path, docker_from: str, tmp_path: Path):
    deb_build_options = {'bldr_foo1', 'bldr_bar2=x'}
    bldr = BLDR(
        local_repo_dir=local_repo_dir,
        source_dir=asset_dir,
        docker_from=docker_from,
        deb_build_options=' '.join(deb_build_options),
    )
    pkg = 'embed-build-opts'
    bldr.build(asset_dir.joinpath(pkg))

    deb_file = list(local_repo_dir.glob('**/' + pkg + '*.deb'))[0]

    extract_dir = tmp_path.joinpath('extracted')
    extract_dir.mkdir()
    extract_deb(deb_file, extract_dir)

    content = extract_dir.joinpath('usr', 'share', 'doc', pkg, 'deb_build_options.txt').read_text()
    assert deb_build_options.issubset(set(content.split())), "The set of build options should not shrink."



import subprocess
from pathlib import Path
from typing import Mapping

import pytest

from bldr.bldr import BLDR
from ..testutil import copytree, extract_deb

PACKAGE_NAME = 'quilt-no-patch'


def git_import_package(git_dir: Path, git_env: Mapping[str, str], branches_dir: Path):
    subprocess.check_call(['git', 'checkout', '-b', 'upstream'], cwd=git_dir)

    copytree(branches_dir.joinpath('upstream'), git_dir)
    subprocess.check_call(['git', 'add', '--all'], cwd=git_dir)
    subprocess.check_call(['git', 'commit', '--no-verify', '--message', 'Imported upstream'], cwd=git_dir, env=git_env)

    subprocess.check_call(['git', 'checkout', '-b', 'ubuntu'], cwd=git_dir)
    copytree(branches_dir.joinpath('debian'), git_dir)
    subprocess.check_call(['git', 'add', '--all'], cwd=git_dir)
    subprocess.check_call(['git', 'commit', '--no-verify', '--message', 'Imported debian'], cwd=git_dir, env=git_env)

    subprocess.check_call(['git', 'checkout', '-b', 'master'], cwd=git_dir)
    copytree(branches_dir.joinpath('master'), git_dir, exist_ok=True)
    subprocess.check_call(['git', 'add', '--all'], cwd=git_dir)
    subprocess.check_call(['git', 'commit', '--no-verify', '--message', 'Our changes'], cwd=git_dir, env=git_env)


@pytest.fixture
def quilt_project_path(tmp_path: Path, asset_dir: Path, git_env: Mapping[str, str]) -> Path:
    quilt_project_dir = tmp_path.joinpath('quilt_project')
    quilt_project_dir.mkdir()
    subprocess.check_call(['git', 'init'], cwd=quilt_project_dir)
    git_import_package(quilt_project_dir, git_env, asset_dir.joinpath(PACKAGE_NAME))
    return quilt_project_dir


def test_quilt_project_build(local_repo_dir: Path, quilt_project_path: Path, docker_from: str, tmp_path: Path):
    bldr = BLDR(
        local_repo_dir=local_repo_dir,
        source_dir=quilt_project_path.parent,
        docker_from=docker_from,
    )
    bldr.build(quilt_project_path)

    quilt_proj_deb_file = list(local_repo_dir.glob('**/' + PACKAGE_NAME + '*.deb'))[0]

    extract_dir = tmp_path.joinpath('extracted')
    extract_dir.mkdir()
    extract_deb(quilt_proj_deb_file, extract_dir)

    content = extract_dir.joinpath('usr', 'share', 'doc', PACKAGE_NAME, 'README').read_text()
    assert content == "Hello patched world!\n", "The patched file should be correct."



import getpass
import io
import os
from pathlib import Path

from pytest_console_scripts import ScriptRunner


def test_simple_commands(script_runner: ScriptRunner, docker_from: str, local_repo_dir: Path):
    stdin = io.StringIO('true\n')
    ret = script_runner.run('bldr', 'shell', docker_from, '--local-repo-dir', local_repo_dir, stdin=stdin)
    assert ret.success, "`true` command should be succeded"
    assert ret.stderr == '', "nothing should be written to stderr"

    stdin = io.StringIO('false\n')
    ret = script_runner.run('bldr', 'shell', docker_from, '--local-repo-dir', local_repo_dir, stdin=stdin)
    assert not ret.success, "`false` command should be failed"
    assert ret.stderr == '', "nothing should be written to stderr"

    stdin = io.StringIO('exit 255\n')
    ret = script_runner.run('bldr', 'shell', docker_from, '--local-repo-dir', local_repo_dir, stdin=stdin)
    assert ret.returncode == 255, "exit code inside the container should be forwarded to the cli"
    assert ret.stderr == '', "nothing should be written to stderr"


def test_who_am_i_should_the_outside_user(script_runner: ScriptRunner, docker_from: str, local_repo_dir: Path):
    stdin = io.StringIO('whoami\n')
    ret = script_runner.run('bldr', 'shell', docker_from, '--local-repo-dir', local_repo_dir, stdin=stdin)
    assert ret.success, "`whoami` command should be succeded"
    assert getpass.getuser() == ret.stdout.strip().split('\n')[-1], "outside user should be inside"
    assert ret.stderr == '', "nothing should be written to stderr"


def test_build_dependencies_should_be_installed(script_runner: ScriptRunner, asset_dir: Path, docker_from: str, local_repo_dir: Path):
    stdin = io.StringIO('dpkg -l libblkid1\n')
    ret = script_runner.run('bldr', 'shell', docker_from, '--local-repo-dir', local_repo_dir, stdin=stdin, cwd=asset_dir.joinpath('example'))
    assert ret.success, "libblkid1 should be installed inside the shell, because it is a build dependency of e2fsprogs."
    assert ret.stderr == '', "nothing should be written to stderr"


def test_http_proxy(script_runner: ScriptRunner, docker_from: str, local_repo_dir: Path):
    env = dict(os.environ)
    env.update({'http_proxy': 'http://proxyurl.example:1234'})
    stdin = io.StringIO('env\n')
    ret = script_runner.run('bldr', 'shell', docker_from, '--local-repo-dir', local_repo_dir, stdin=stdin, env=env)
    assert ret.success, "`env` command should be succeded"
    assert 'http_proxy=http://proxyurl.example:1234' in ret.stdout, "The shell command should pass the http_proxy env var"
    assert ret.stderr == '', "nothing should be written to stderr"

    stdin = io.StringIO('sudo env\n')
    ret = script_runner.run('bldr', 'shell', docker_from, '--local-repo-dir', local_repo_dir, stdin=stdin, env=env)
    assert ret.success, "`env` command as supersuer should be succeded"
    assert 'http_proxy=http://proxyurl.example:1234' in ret.stdout, "The env var 'http_proxy' should survive 'sudo'"
    assert ret.stderr == '', "nothing should be written to stderr"



import subprocess
from pathlib import Path
from typing import Mapping

import pytest

from bldr.bldr import BLDR
from ..testutil import copytree, extract_deb


@pytest.fixture
def quilt_project_path(tmp_path: Path, asset_dir: Path, git_env: Mapping[str, str]) -> Path:
    quilt_project_dir = tmp_path.joinpath('quilt_project')
    quilt_project_dir.mkdir()
    subprocess.check_call(['git', 'init'], cwd=quilt_project_dir)
    subprocess.check_call(['git', 'checkout', '-b', 'upstream'], cwd=quilt_project_dir)

    copytree(asset_dir.joinpath('test-quilt-proj-onedir', 'upstream'), quilt_project_dir)
    subprocess.check_call(['git', 'add', '--all'], cwd=quilt_project_dir)
    subprocess.check_call(['git', 'commit', '--no-verify', '--message', 'Imported upstream'], cwd=quilt_project_dir, env=git_env)

    subprocess.check_call(['git', 'checkout', '-b', 'ubuntu'], cwd=quilt_project_dir)
    copytree(asset_dir.joinpath('test-quilt-proj-onedir', 'debian'), quilt_project_dir)
    subprocess.check_call(['git', 'add', '--all'], cwd=quilt_project_dir)
    subprocess.check_call(['git', 'commit', '--no-verify', '--message', 'Imported debian'], cwd=quilt_project_dir, env=git_env)

    subprocess.check_call(['git', 'checkout', '-b', 'master'], cwd=quilt_project_dir)
    subprocess.check_call(['git', 'reset', '--hard', 'ubuntu'], cwd=quilt_project_dir)

    return quilt_project_dir


def test_quilt_project_build(local_repo_dir: Path, quilt_project_path: Path, docker_from: str, tmp_path: Path):
    bldr = BLDR(
        local_repo_dir=local_repo_dir,
        source_dir=quilt_project_path.parent,
        docker_from=docker_from,
    )
    bldr.build(quilt_project_path)

    quilt_proj_deb_file = list(local_repo_dir.glob('**/quilt-proj*.deb'))[0]

    extract_dir = tmp_path.joinpath('extracted')
    extract_dir.mkdir()
    extract_deb(quilt_proj_deb_file, extract_dir)

    content = extract_dir.joinpath('usr', 'share', 'doc', 'feeling', 'alone').read_text()
    assert content == "Hello, friend!\n", "The patched file should be correct."

    content = extract_dir.joinpath('usr', 'share', 'doc', 'feeling', 'lonely').read_text()
    assert content == "I am a test.\n", "The patched file should be correct."



import pytest
import shutil
from pathlib import Path

DEFAULT_DOCKER_IMAGES = ["ubuntu:xenial", "ubuntu:bionic", "ubuntu:focal", "debian:bullseye"]


def pytest_addoption(parser):
    parser.addoption(
        "--docker-image", nargs="+", default=DEFAULT_DOCKER_IMAGES, help="Select docker image to run the tests on (default: %(default)s)",
    )


def pytest_generate_tests(metafunc):
    if "docker_from" in metafunc.fixturenames:
        docker_image = metafunc.config.getoption("docker_image")
        metafunc.parametrize("docker_from", docker_image)


@pytest.fixture
def asset_dir() -> Path:
    test_dir = Path(__file__).resolve().parent
    return test_dir.joinpath('asset')


@pytest.fixture
def local_repo_dir(tmp_path: Path) -> Path:
    return tmp_path.joinpath('local_repo_dir')


@pytest.fixture
def reindex_data(tmp_path: Path, asset_dir: Path) -> Path:
    reindex_data_dir = tmp_path.joinpath('test-reindex-data')
    shutil.copytree(
        asset_dir.joinpath('test-reindex-data'),
        reindex_data_dir,
    )

    return reindex_data_dir



#!/usr/bin/env python3

import glob
import os
from pathlib import Path
from setuptools import setup, find_packages

from bldr.version import get_version


def get_package_data():
    package_data = []
    for path in glob.glob('bldr/data/**/*', recursive=True):
        if os.path.isfile(path):
            package_data.append(path.split('/', 1)[1])
    return package_data


with open("README.md") as readme:
    LONG_DESCRIPTION = readme.read()

setup(
    name='bldr',
    author_email="noreply@oneidentity.com",
    author="BLDR Developers",
    url="https://www.github.com/OneIdentity/bldr",
    description="Build debian packages in a clean docker environment",
    long_description=LONG_DESCRIPTION,
    long_description_content_type="text/markdown",
    version=get_version(),
    packages=find_packages(include=["bldr*"]),
    install_requires=Path('requirements.txt').read_text(),
    entry_points={
        "console_scripts": [
            "bldr = bldr.cli:main",
        ]
    },
    extras_require={
        'dev': Path('requirements-dev.txt').read_text(),
    },
    python_requires=">=3.6",
    package_data={'bldr': [
        'VERSION',
    ] + get_package_data()},
    license="MIT",
    classifiers=[
        "Development Status :: 4 - Beta",
        "Intended Audience :: System Administrators",
        "License :: OSI Approved :: MIT License",
        "Operating System :: POSIX :: Linux",
        "Programming Language :: Python",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Topic :: System :: Systems Administration",
        "Topic :: Utilities",
    ],
)



import argparse
import json
import logging

from abc import abstractmethod, ABC
from pathlib import Path
from typing import Any, Dict, IO, List, Optional, Sequence, Tuple


class ConfigLoaderError(Exception):
    pass


class ConfigLoader(ABC):
    def __init__(self, paths: List[Path]):
        self.paths = paths
        self.config = None

    def copy(self) -> 'ConfigLoader':
        return type(self)(list(self.paths))

    def load_config_file(self, path: Path) -> Dict:
        with path.open() as config_file:
            config = self.parse(config_file)
        return config

    def load(self) -> Dict[str, Any]:
        retval: Dict[str, Any] = {}
        for path in self.paths:
            try:
                config = self.load_config_file(path)
            except IOError:
                continue

            for key, value in config.items():
                if value is None:
                    if key in retval:
                        retval.pop(key)
                else:
                    retval[key] = value

        return retval

    @abstractmethod
    def parse(self, file_object: IO[str]) -> Dict[str, Any]:
        pass


class JSONConfigLoader(ConfigLoader):
    def parse(self, file_object: IO[str]) -> Dict[str, Any]:
        return json.load(file_object)


def parse_action(action: argparse.Action, args: Optional[Sequence[str]] = None) -> Tuple[argparse.Namespace, List[str]]:
    parser = argparse.ArgumentParser(add_help=False)
    parser._add_action(action)
    namespace, argv = parser.parse_known_args(args)
    return (namespace, argv)


class SubParsers:
    def __init__(self, config_loader, subparsers, required=True):
        self.config_loader = config_loader
        self.subparsers = subparsers
        self.subparsers.required = required

    def add_parser(self, *args, **kwargs):
        kwargs = kwargs.copy()
        kwargs["config_loader"] = self.config_loader
        parser = self.subparsers.add_parser(*args, **kwargs)
        return parser


class ArgumentParser(argparse.ArgumentParser):
    def __init__(self, *args: Any, **kwargs: Any):
        self.config_loader = kwargs.pop("config_loader", None)
        self.logger = logging.getLogger(__name__)

        super().__init__(*args, **kwargs)

    def parse_known_args(self, args: Optional[Sequence[str]] = None, namespace: argparse.Namespace = None) -> Tuple[argparse.Namespace, List[str]]:
        if self.config_loader:
            self.set_defaults_from_config(args)
        return super().parse_known_args(args, namespace=None)

    def set_defaults_from_config(self, args: Optional[Sequence[str]] = None) -> None:
        # make a copy here as we may adjust the paths attribute below
        loader = self.config_loader.copy()

        config = loader.load()

        self.set_defaults(**config)



from bldr.utils import get_config_file_paths
import shutil
import subprocess
from pathlib import Path
import pytest


def copytree(source: Path, destination: Path, exist_ok: bool = False) -> None:
    for source_child in source.iterdir():
        destination_child = destination.joinpath(source_child.name)
        if source_child.is_dir():
            destination_child.mkdir(exist_ok=exist_ok)
            copytree(source_child, destination_child, exist_ok=exist_ok)
        else:
            shutil.copy2(str(source_child), str(destination_child))


def extract_deb(deb_file: Path, extract_dir: Path) -> None:
    subprocess.check_call(['ar', 'x', deb_file], cwd=extract_dir)
    tar_file = list(extract_dir.glob('data.tar.*'))[0]
    subprocess.check_call(['tar', '-xf', tar_file], cwd=extract_dir)


def get_existing_config_paths():
    return [str(path) for path in get_config_file_paths() if path.exists()]


skip_if_global_config = pytest.mark.skipif(
    len(get_existing_config_paths()) > 0,
    reason="Config file(s) already exist(s): {}".format(", ".join(get_existing_config_paths()))
)



#
# Copyright (c) 2006-2020 Balabit
# All Rights Reserved.
#

import json
import pytest

from pathlib import Path
from bldr.cli import CLI
from ..testutil import skip_if_global_config


def test_config_with_an_existent_file(tmp_path: Path):
    config = tmp_path.joinpath('config.json')
    with config.open('w') as config_file:
        json.dump({'container_env': ['foo=bar', 'bar=baz']}, config_file)
    bldr = CLI(['bldr', '--config', str(config), 'build', 'ubuntu:bionic'])
    assert bldr.args.container_env == [('foo', 'bar'), ('bar', 'baz')]


def test_config_and_cmdline_arg_has_a_same_value_and_a_proper_type(tmp_path: Path):
    config = tmp_path.joinpath('config.json')
    with config.open('w') as config_file:
        json.dump({'hooks_dir': '/foo/hooks'}, config_file)
    bldr = CLI(['bldr', '--config', str(config), 'build', 'ubuntu:bionic'])
    hooks_dir1 = bldr.args.hooks_dir
    bldr = CLI(['bldr', 'build', 'ubuntu:bionic', '--hooks-dir', '/foo/hooks'])
    hooks_dir2 = bldr.args.hooks_dir

    assert hooks_dir1 == Path('/foo/hooks')
    assert hooks_dir2 == Path('/foo/hooks')


def test_config_override_with_cmdline_arg(tmp_path):
    config = tmp_path.joinpath('config.json')
    with config.open('w') as config_file:
        json.dump({'hooks_dir': '/foo/hooks'}, config_file)
    bldr = CLI(
        ['bldr', 'build', 'ubuntu:bionic', '--hooks-dir', '/override/hooks']
    )

    assert bldr.args.hooks_dir == Path('/override/hooks')


@skip_if_global_config
@pytest.mark.parametrize(
    'config_data, arguments, expected',
    [
        (
            {'container_env': ['foo=bar']},
            [],
            [('foo', 'bar')],
        ),
        (
            {'container_env': ['foo=bar', 'bar=baz']},
            [],
            [('foo', 'bar'), ('bar', 'baz')],
        ),
        (
            {'container_env': ['foo=bar']},
            ['--container-env', 'bar=baz'],
            [('foo', 'bar'), ('bar', 'baz')],
        ),
        (
            {},
            ['--container-env', 'foo=bar'],
            [('foo', 'bar')]
        ),
        (
            {},
            ['--container-env', 'foo=bar', '--container-env', 'bar=baz'],
            [('foo', 'bar'), ('bar', 'baz')],
        ),
    ]
)
def test_container_env(tmp_path: Path, config_data, arguments, expected):
    config = tmp_path.joinpath('config.json')

    with config.open('w') as config_file:
        json.dump(config_data, config_file)
    base_arguments = [
        'bldr', '--config', str(config), 'build', 'ubuntu:bionic',
    ]
    bldr = CLI(base_arguments + arguments)

    assert bldr.args.container_env == expected
